/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2019  Dávid Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org
*/

#include "common.h"
#include <string.h>

// data:3D1A
sbyte distance_mirror;

// custom logic - prison guard on level, room 18 looks different
byte* prison_guard_palette = (byte[]) {
	0x00, 0x00, 0x00,
	      0x2A, 0x0C, 0x00,
	      0x20, 0x09, 0x00,
	      0x3F, 0x1E, 0x1E,
	      0xFF, 0x00, 0x00,
	      0x12, 0x06, 0x06,
	      0x18, 0x10, 0x10,
	      0x21, 0x1A, 0x30,
	      0x1D, 0x17, 0x2A,
	      0x2C, 0x22, 0x3F,
	      0x1E, 0x14, 0x14,
	      0xFF, 0xFF, 0x00,
	      0xFF, 0xFF, 0x00,
	      0x20, 0x09, 0x00,
	      0xFF, 0xFF, 0x00,
	      0x3D, 0x00, 0x0E
};

// seg003:0000
void __pascal far init_game(int level) {
	if(offscreen_surface) {
		free_surface(offscreen_surface); // missing in original
		offscreen_surface = NULL;
	}
	offscreen_surface = make_offscreen_buffer(&rect_top);
	load_kid_sprite();
	text_time_remaining = 0;
	text_time_total = 0;
	is_show_time = 0;
	enable_lighting = 0;
	checkpoint = 0;
	// custom logic - reset checkpoint 2
	checkpoint2 = 0;
	// custom logic - checkpoint must not be activated
	custom->checkpoint_level = 0;
	// custom logic - kid exited from the left door on level 2
	if (level != 3) {
		level_2_right_door = 0;
	}
	// custom logic - falling checkpoint on level 6
	if (level != 6) {
		falling_exit_checkpoint = 0;
	}
	// custom logic - kid's default direction at the end of level 6 is to the left
	if (level != 7) {
		falling_exit_final_direction = dir_FF_left;
	}
	//upside_down = 0; // N.B. upside_down is also reset in set_start_pos()
	resurrect_time = 0;
	if (!dont_reset_time) {
		rem_min = custom->start_minutes_left;   // 60
		rem_tick = custom->start_ticks_left;    // 719
		hitp_beg_lev = custom->start_hitp;      // 3
	}
	need_level1_music = (level == /*1*/ custom->intro_music_level);
	play_level(level);
}

// seg003:005C
void __pascal far play_level(int level_number) {
	cutscene_ptr_type cutscene_func;
#ifdef USE_COPYPROT
	if (enable_copyprot && level_number == custom->copyprot_level) {
		level_number = 15;
	}
#endif
	for (;;) {
		enable_lighting = ((level_number == 6 && falling_exit_checkpoint >= 4) ||
							(level_number >= 7 && level_number <= 9))
			? 1
			: 0;
		if (demo_mode && level_number > 2) {
			start_level = -1;
			need_quotes = 1;
			start_game();
		}
		if (level_number != current_level) {
			if (level_number < 0 || level_number > 15) {
				printf("Tried to load cutscene for level %d, not in 0..15\n", level_number);
				quit(1);
			}
			cutscene_func = tbl_cutscenes[custom->tbl_cutscenes_by_index[level_number]];
			if (level_number == 14 && initial_mouse_timer == -2 && cutscene_func == NULL) {
				cutscene_func = tbl_cutscenes[9];
			}
			if (cutscene_func != NULL

#ifdef USE_REPLAY
			        && !(recording || replaying)
#endif
#ifdef USE_SCREENSHOT
			        && !want_auto_screenshot()
#endif
			   ) {
				load_intro(level_number > 2, cutscene_func, 1);
			}
		}
		load_lev_spr(level_number);
		load_level();
		pos_guards();
		clear_coll_rooms();
		clear_saved_ctrl();
		drawn_room = 0;
		mobs_count = 0;
		trobs_count = 0;
		next_sound = -1;
		holding_sword = 0;
		grab_timer = 0;
		can_guard_see_kid = 0;
		united_with_shadow = 0;
		flash_time = 0;
		leveldoor_open = 0;
		// custom logic - makes shadow blink on level 9
		kid_shadow_blink_timer = 0;
		// custom logic - tracking whether the kid met Jaffar
		jaffar_presented = false;
		// custom logic - take 2 points per hit from Jaffar
		jaffar_double_hit = false;
		// custom logic - tracking whether the kid met Jaffar's vision
		jaffar_vision_presented = false;
		// custom logic - tracking antidote on level 5
		took_antidote = 0;
		// custom logic - tracking shadow crashing the floor
		level_7_tile_crashed = false;
		// custom logic - timer that kills kid on level 12B when there is nowhere to run
		level_13_nowhere_to_run_timer = -1;
		// custom logic - tracks whether shadow appeared on this level
		shadow_appeared_on_level = 0;
		// custom logic - an ability to restart the level on a timer
		level_restart_timer = 0;
		// custom logic - tracking whether the level was started upside down
		level_start_upside_down = upside_down;
		// custom logic - guard is alive on a debris tile
		guard_alive_on_debris = 0;
		// custom logic - set blink timer to 0
		blink_timer = 0;
		// custom logic - reset number of potions shadow drank on lower levels
		if (level_number < 7) {
			initial_shadow_potion_counter = 0;
			shadow_potion_counter = 0;
		} else if (level_number < 10 && shadow_potion_counter == 0xFFFF) {
			initial_shadow_potion_counter = 0;
			shadow_potion_counter = 0;
		} else {
			shadow_potion_counter = initial_shadow_potion_counter;
		}
		// custom logic - reset initial mouse timer
		if (level_number <= 12) {
			initial_mouse_timer = 0;
		}
		// custom logic - tracking last drank potion
		potion_room = 0;
		potion_pos = -1;
		potion_counter = 0;
		potion_size = 0;
		// custom logic - tracking kid standing near a large blue potion
		kid_near_big_blue_potion = -1;
		demo_index = 0;
		demo_time = 0;
		guardhp_curr = 0;
		hitp_delta = 0;
		Guard.charid = charid_2_guard;
		Guard.direction = dir_56_none;
		do_startpos();
		// custom logic - no sword on level 15
		have_sword = /*(level_number != 1)*/ ((level_number == 0 || level_number >= custom->have_sword_from_level) && level_number != 15);
		find_start_level_door();
		// busy waiting?
		while (check_sound_playing() && !do_paused()) idle();
		stop_sounds();
#ifdef USE_REPLAY
		if (replaying) replay_restore_level();
		if (skipping_replay) {
			if (replay_seek_target == replay_seek_0_next_room ||
			        replay_seek_target == replay_seek_1_next_level
			   )
				skipping_replay = 0; // resume replay from here
		}
#endif
		draw_level_first();
		show_copyprot(0);
		level_number = play_level_2();
		free_peels();
	}
}

// seg003:01A3
void __pascal far do_startpos() {
	word x;
	// custom logic - set the wall patern palette here because it gets overridden for some levels
	if (custom->tbl_level_type[current_level] == 1 /* palace */) {
		// (palace wall pattern) #C09850 = light brown
		set_pal(6, 0x30, 0x26, 0x14, 0);
	}
	// Special event: start at checkpoint
	if (current_level == 3) {
		if (checkpoint) {
			level.start_dir = /*dir_FF_left*/ custom->checkpoint_respawn_dir;
			level.start_room = /*2*/ custom->checkpoint_respawn_room;
			level.start_pos = /*6*/ custom->checkpoint_respawn_tilepos;
		} else if (level_2_right_door == 0) {
			level.start_dir = dir_FF_left;
			level.start_room = 4;
			level.start_pos = 25;
		}
	// custom logic - configuring starting position for level 6B
	} else if (current_level == 6 && falling_exit_checkpoint >= 4) {
		level.start_dir = dir_0_right;
		level.start_room = 12;
		level.start_pos = 22;
		load_lev6_palette();
		set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
		play_sound(sound_1_falling);
	// custom logic - configuring starting position for level 7
	} else if (current_level == 7) {
		level.start_dir = falling_exit_final_direction;
		if (level.start_dir == dir_0_right) {
			++level.start_pos;
		}
	// custom logic - levels 10 and 11 have a green wall palette
	} else if (current_level == 10 || current_level == 11) {
		set_pal(6, 0x16, 0x20, 0x16, 0);
	// custom logic - configuring starting position for level 12
	// after the checkpoint
	} else if (current_level == 12 && checkpoint2) {
		level.start_dir = dir_FF_left;
		level.start_room = 12;
		level.start_pos = 19;
	}
	next_room = Char.room = level.start_room;
	x = level.start_pos;
	Char.curr_col = x % 10;
	Char.curr_row = x / 10;
	Char.x = x_bump[Char.curr_col + 5] + 14;
	// Start in the opposite direction (and turn into the correct one).
	Char.direction = ~ level.start_dir;
	if (seamless == 0) {
		if (current_level != 0) {
			x = hitp_beg_lev;
			// custom logic kid looses 1 hit point from jumping through the mirror
			if (current_level == 6 && falling_exit_checkpoint >= 4) {
				--x;
			}
		} else {
			// HP on demo level
			x = /*4*/ custom->demo_hitp;
		}
		hitp_max = hitp_curr = x;
	}
	if (/*current_level == 1*/ custom->tbl_entry_pose[current_level] == 1) {
		// Special event: press button + falling entry
		get_tile(5, 2, 0);
		trigger_button(0, 0, -1);
		seqtbl_offset_char(seq_7_fall); // fall
	} else if (/*current_level == 13*/ custom->tbl_entry_pose[current_level] == 2) {
		// Special event: running entry
		seqtbl_offset_char(seq_84_run); // run
	// custom logic - reset the fall on level 6B
	} else if (current_level == 6 && falling_exit_checkpoint >= 4) {
		seqtbl_offset_char(seq_2_stand); // reset the fall
	} else {
		seqtbl_offset_char(seq_5_turn); // turn
	}
	// custom logic - start drinking the potion on level 6
	if (current_level == 6) {
		if (falling_exit_checkpoint >= 4) {
			is_blind_mode = 0;
		} else {
			is_blind_mode = 1;
			draw_rect(&rect_top, 0);
			potion_room = 24;
			potion_pos = get_tilepos(5, 1);
			potion_counter = 0;
			potion_size = 1;
			seqtbl_offset_char(seq_78_drink); // drink
			// the empty potion is in a large pot
			if (get_tile(6, 6, 1) == tiles_10_potion) {
				curr_room_modif[curr_tilepos] = 0x40;
			}
		}
	} else {
		is_blind_mode = 0;
	}
	// custom logic - the gate in the escaped cell should be partially opened
	if (current_level == 1) {
		if (get_tile(18, 3, 0) == tiles_4_gate) {
			curr_room_modif[curr_tilepos] = 85;
		}
	// custom logic - randomize blue potions on level 8, room 19
	} else if (current_level == 8) {
		prandom(1); // initialize the seed, just in case
		int door_opener_potion_tile = prandom(4);
		if (get_tile(19, 5, 0) == tiles_10_potion) {
			curr_room_modif[curr_tilepos] = (door_opener_potion_tile == 0 ? 0x30 : 0x28);
		}
		if (get_tile(19, 3, 1) == tiles_10_potion) {
			curr_room_modif[curr_tilepos] = (door_opener_potion_tile == 1 ? 0x30 : 0x28);
		}
		if (get_tile(19, 5, 1) == tiles_10_potion) {
			curr_room_modif[curr_tilepos] = (door_opener_potion_tile == 2 ? 0x30 : 0x28);
		}
		if (get_tile(19, 7, 1) == tiles_10_potion) {
			curr_room_modif[curr_tilepos] = (door_opener_potion_tile == 3 ? 0x30 : 0x28);
		}
		if (get_tile(19, 8, 2) == tiles_10_potion) {
			curr_room_modif[curr_tilepos] = (door_opener_potion_tile == 4 ? 0x30 : 0x28);
		}
	// custom logic - the level opener door is stuck when the kid has not joined with shadow
	} else if (current_level == 10) {
		get_tile(8, 1, 0);
		curr_room_tiles[curr_tilepos] = tiles_5_stuck;
	// custom logic - remove sword if kid had joined shadow
	} else if (current_level == 12) {
		if (initial_shadow_potion_counter == 0xFFFF && get_tile(15, 1, 0) == tiles_22_sword) {
			curr_room_tiles[curr_tilepos] = tiles_1_floor;
		}
	// custom logic - the level door on level 12B should appear to be open
	} else if (current_level == 13) {
		if (get_tile(12, 3, 1) == tiles_16_level_door_left) {
			curr_room_modif[curr_tilepos] = 43;
		}
	}

	set_start_pos();
}

// seg003:028A
void __pascal far set_start_pos() {
	Char.y = y_land[Char.curr_row + 1];
	Char.alive = -1;
	Char.charid = charid_0_kid;
	is_screaming = 0;
	knock = 0;
	// custom logic - remember the upside down mode between levels 2 and 14
	if (current_level < 2 || current_level > 14) {
		upside_down = 0;
	}
	is_feather_fall = 0;
	Char.fall_y = 0;
	Char.fall_x = 0;
	offguard = 0;
	Char.sword = sword_0_sheathed;
	droppedout = 0;
	play_seq();
	// custom logic - do not show room 12 when the prince is falling into level 6B
	if (current_level == 6 && Char.room == 12 && falling_exit_checkpoint >= 4) {
		goto_other_room(3);
	} else if (current_level == /*7*/ custom->falling_entry_level && Char.room == /*17*/ custom->falling_entry_room) {
		// Special event: level 7 falling entry
		// level 7, room 17: show room below
		goto_other_room(3);
	}
	savekid();
}

// seg003:02E6
void __pascal far find_start_level_door() {
	short tilepos;
	get_room_address(Kid.room);
	for (tilepos = 0; tilepos < 30; ++tilepos) {
		if ((curr_room_tiles[tilepos] & 0x1F) == tiles_16_level_door_left) {
			start_level_door(Kid.room, tilepos);
		}
	}
}

// seg003:0326
void __pascal far draw_level_first() {
	next_room = Kid.room;
	check_the_end();
	if (custom->tbl_level_type[current_level]) {
		gen_palace_wall_colors();
	}
	draw_rect(&screen_rect, 0);
	show_level();
	redraw_screen(0);
	draw_kid_hp(hitp_curr, hitp_max);

#ifdef USE_QUICKSAVE
	check_quick_op();
#endif

#ifdef USE_SCREENSHOT
	auto_screenshot();
#endif

	// Busy waiting!
	start_timer(timer_1, 5);
	do_simple_wait(1);
}

// seg003:037B
void __pascal far redraw_screen(int drawing_different_room) {
	//remove_flash();
	if (drawing_different_room) {
		draw_rect(&rect_top, 0);
	}

	different_room = 0;
	if (is_blind_mode) {
		draw_rect(&rect_top, 0);
	} else {
		// Moved *before* drawings.
		if (curr_guard_color) {
			// custom logic - prison guard on level 1, room 18 looks different
			// level 15 has the same guard
			if ((current_level == 1 && drawn_room == 18) || current_level == 15) {
				set_chtab_palette(chtab_addrs[id_chtab_5_guard], prison_guard_palette, 0x10);
			} else {
				set_chtab_palette(chtab_addrs[id_chtab_5_guard], &guard_palettes[0x30 * curr_guard_color - 0x30], 0x10);
			}
		}
		need_drects = 0;
		redraw_room();
#ifdef USE_LIGHTING
		redraw_lighting();
#endif
		if (is_keyboard_mode) {
			clear_kbd_buf();
		}
		is_blind_mode = 1;
		draw_tables();
		if (is_keyboard_mode) {
			clear_kbd_buf();
		}
#ifdef USE_COPYPROT
		if (current_level == 15) {
			// letters on potions level
			current_target_surface = offscreen_surface;
			short var_2;
			for (var_2 = 0; var_2 < 14; ++var_2) {
				if (copyprot_room[var_2] == drawn_room) {
					set_curr_pos((copyprot_tile[var_2] % 10 << 5) + 24, copyprot_tile[var_2] / 10 * 63 + 38);
					draw_text_character(copyprot_letter[cplevel_entr[var_2]]);
				}
			}
			current_target_surface = onscreen_surface_;
		}
#endif
		is_blind_mode = 0;
		memset_near(table_counts, 0, sizeof(table_counts));
		draw_moving();
		draw_tables();
		if (is_keyboard_mode) {
			clear_kbd_buf();
		}
		need_drects = 1;
		if (upside_down) {
			flip_screen(offscreen_surface);
		}
		copy_screen_rect(&rect_top);
		if (upside_down) {
			flip_screen(offscreen_surface);
		}
		if (is_keyboard_mode) {
			clear_kbd_buf();
		}
	}
	exit_room_timer = 2;

}

#ifdef CHECK_TIMING
typedef struct test_timing_state_type {
	bool already_had_first_frame;
	Uint64 level_start_counter;
	int ticks_left_at_level_start;
	float seconds_left_at_level_start;
} test_timing_state_type;

void test_timings(test_timing_state_type* state) {

	if (!state->already_had_first_frame) {
		state->level_start_counter = SDL_GetPerformanceCounter();
		state->ticks_left_at_level_start = (rem_min-1)*720 + rem_tick;
		state->seconds_left_at_level_start = (1.0f / 60.0f) * (5 * state->ticks_left_at_level_start);
		printf("Seconds left = %f\n", state->seconds_left_at_level_start);
		state->already_had_first_frame = true;
	} else if (rem_tick % 12 == 11) {
		Uint64 current_counter = SDL_GetPerformanceCounter();
		float actual_seconds_elapsed = (float)(current_counter - state->level_start_counter) / (float)SDL_GetPerformanceFrequency();

		int ticks_left = (rem_min-1)*720 + rem_tick;
		float game_seconds_left = (1.0f / 60.0f) * (5 * ticks_left);
		float game_seconds_elapsed = state->seconds_left_at_level_start - game_seconds_left;

		printf("rem_min: %d   game elapsed (s): %.2f    actual elapsed (s): %.2f     delta: %.2f\n",
		       rem_min, game_seconds_elapsed, actual_seconds_elapsed, actual_seconds_elapsed - game_seconds_elapsed);
	}

}
#endif

// seg003:04F8
// Returns a level number:
// - The current level if it was restarted.
// - The next level if the level was completed.
int __pascal far play_level_2() {
	reset_timer(timer_1);
#ifdef CHECK_TIMING
	test_timing_state_type test_timing_state = {0};
#endif
	while (1) { // main loop
#ifdef USE_QUICKSAVE
		check_quick_op();
#endif
#ifdef CHECK_TIMING
		test_timings(&test_timing_state);
#endif

#ifdef USE_REPLAY
		if (need_replay_cycle) replay_cycle();
#endif
		// custom logic - level 6 feels faster (prince is drunk) in rooms
		// before the mirror and slower in rooms after the mirror.
		if (current_level == 6 && Kid.room != 1) {
			if (Kid.room != 3 &&
			        Kid.room != 7 &&
			        Kid.room != 17 &&
			        Kid.room != 19 &&
			        Kid.room != 20 &&
			        Kid.room != 21 &&
			        Kid.room != 22) {
				// speed when not fighting (smaller is faster)
				set_timer_length(timer_1, 3);
			} else {
				// speed when fighting (smaller is faster)
				if (Kid.sword == sword_2_drawn) {
					// speed when fighting (smaller is faster)
					set_timer_length(timer_1, 7);
				} else {
					// speed when not fighting (smaller is faster)
					set_timer_length(timer_1, 6);
				}
			}
			// custom logic - kid is slower with only 2 hit points
		} else if (hitp_curr == 2) {
			// speed when fighting (smaller is faster)
			if (Kid.sword == sword_2_drawn) {
				// speed when fighting (smaller is faster)
				set_timer_length(timer_1, 7);
			} else {
				// speed when not fighting (smaller is faster)
				set_timer_length(timer_1, 6);
			}
			// custom logic - kid is sluggish with a single hit point
		} else if (hitp_curr < 2) {
			// speed when fighting (smaller is faster)
			if (Kid.sword == sword_2_drawn) {
				// speed when fighting (smaller is faster)
				set_timer_length(timer_1, 8);
			} else {
				// speed when not fighting (smaller is faster)
				set_timer_length(timer_1, 7);
			}
		} else {
			if (Kid.sword == sword_2_drawn) {
				// speed when fighting (smaller is faster)
				set_timer_length(timer_1, 6);
			} else {
				// speed when not fighting (smaller is faster)
				set_timer_length(timer_1, 5);
			}
		}
		guardhp_delta = 0;
		hitp_delta = 0;
		timers();
		play_frame();

#ifdef USE_REPLAY
		// At the exact "end of level" frame, preserve the seed to ensure reproducibility,
		// regardless of how long the sound is still playing *after* this frame (torch animation modifies the seed!)
		if (keep_last_seed == 1) {
			preserved_seed = random_seed;
			keep_last_seed = -1; // disable repeat
		}
#endif

		if (is_restart_level) {
			is_restart_level = 0;
			reset_level3_checkpoint_state(); // reset level 3 state
			reset_level12_checkpoint_state(); // custom logic - reset level 12 state after checkpoint
			upside_down = level_start_upside_down;
			return current_level;
		} else {
			if (next_level == current_level || check_sound_playing()) {
				draw_game_frame();
				flash_if_hurt();
				remove_flash_if_hurt();
				do_simple_wait(timer_1);
				reset_level3_checkpoint_state(); // custom logic - reset level 3 state after checkpoint
				reset_level12_checkpoint_state(); // custom logic - reset level 12 state after checkpoint
			} else {
				stop_sounds();
				hitp_beg_lev = hitp_max;
				checkpoint = 0;
				// custom logic - reset checkpoint on level 12
				checkpoint2 = 0;
				// custom logic - falling checkpoint on level 6
				falling_exit_checkpoint = 0;
				// custom logic - checkpoint must not be activated until kid drinks the potion
				custom->checkpoint_level = 0;
				// custom logic - kid's default exit door at the end of level 2 is left
				if (next_level != 3) {
					level_2_right_door = 0;
				}
				// custom logic - setting the initial number of potions shadow drank on the next level
				initial_shadow_potion_counter = shadow_potion_counter;
				// custom logic - kid's default direction at the end of level 6 is to the left
				if (next_level != 7) {
					falling_exit_final_direction = dir_FF_left;
				}

#ifdef USE_REPLAY
				if (keep_last_seed == -1) {
					random_seed = preserved_seed; // Ensure reproducibility in the next level.
					keep_last_seed = 0;
				}
#endif
				if (next_level == 7) {
					auto_save_level(7);
				} else {
					need_auto_save = 1;
				}

				return next_level;
			}
		}
	}
}

// custom logic - on quick load of level 3 enable checkpoint if the white potion is not in place
// and adjust the tiles
void __pascal far reset_level3_checkpoint_state() {
	if (current_level == 3) {
		get_tile(16, 0, 0);
		if (curr_room_tiles[curr_tilepos] != tiles_10_potion  ||
		        curr_room_modif[curr_tilepos] >= 0x38) {
			custom->checkpoint_level = 3;
		} else {
			custom->checkpoint_level = 0;
		}
		if (checkpoint) {
			// custom logic - drop the loose tile onto the closer tile
			get_tile(8, 4, 0);
			curr_room_tiles[curr_tilepos] = tiles_0_empty;
			get_tile(7, 4, 0);
			curr_room_tiles[curr_tilepos] = tiles_14_debris;
			// custom logic - make large red and white potions empty
			get_tile(17, 3, 2);
			curr_room_modif[curr_tilepos] = 0x40;
			get_tile(16, 0, 0);
			curr_room_modif[curr_tilepos] = 0x40;
			// custom logic - remove skeleton where the large potion used to be
			// and put him behind the gate
			get_tile(17, 4, 2);
			curr_room_tiles[curr_tilepos] = tiles_1_floor;
			get_tile(17, 8, 2);
			curr_room_tiles[curr_tilepos] = tiles_21_skeleton;
			// custom logic - remove the top row skeleton and put him on the bottom
			get_tile(14, 2, 0);
			curr_room_tiles[curr_tilepos] = tiles_1_floor;
			get_tile(13, 3, 2);
			curr_room_tiles[curr_tilepos] = tiles_21_skeleton;
			// custom logic - drop the loose tile below the large white potion
			get_tile(15, 5, 2);
			curr_room_tiles[curr_tilepos] = tiles_0_empty;
			get_tile(18, 5, 0);
			curr_room_tiles[curr_tilepos] = tiles_14_debris;
			get_tile(22, 4, 2);
			curr_room_tiles[curr_tilepos] = tiles_14_debris;
		}
	}
}

// custom logic - on quick load of level 12 enable checkpoint
void __pascal far reset_level12_checkpoint_state() {
	if (current_level == 12 && checkpoint2) {
		// drop tile in room 14
		get_tile(14, 2, 2);
		curr_room_tiles[curr_tilepos] = tiles_0_empty;
		redraw_height = 35;
		set_wipe(potion_pos, 1);
		set_redraw_full(potion_pos, 1);
		// add debris in the room below
		get_tile(24, 2, 1);
		curr_room_tiles[curr_tilepos] = tiles_14_debris;
		// drop lower tiles
		get_tile(9, 0, 0);
		curr_room_tiles[curr_tilepos] = tiles_0_empty;
		get_tile(9, 1, 0);
		curr_room_tiles[curr_tilepos] = tiles_0_empty;
		get_tile(1, 9, 0);
		curr_room_tiles[curr_tilepos] = tiles_0_empty;
		get_tile(7, 6, 1);
		curr_room_tiles[curr_tilepos] = tiles_0_empty;
		get_tile(8, 2, 1);
		curr_room_tiles[curr_tilepos] = tiles_0_empty;
		// add debris in the room below
		get_tile(3, 6, 2);
		curr_room_tiles[curr_tilepos] = tiles_14_debris;
		get_tile(3, 9, 2);
		curr_room_tiles[curr_tilepos] = tiles_14_debris;
		get_tile(4, 0, 2);
		curr_room_tiles[curr_tilepos] = tiles_14_debris;
		get_tile(4, 1, 2);
		curr_room_tiles[curr_tilepos] = tiles_14_debris;
		// move skeleton tile
		get_tile(24, 7, 1);
		curr_room_tiles[curr_tilepos] = tiles_1_floor;
		get_tile(4, 6, 2);
		curr_room_tiles[curr_tilepos] = tiles_21_skeleton;
	}
}

// seg003:0576
void __pascal far redraw_at_char() {
	short x_top_row;
	short tile_col;
	short tile_row;
	short x_col_left;
	short x_col_right;
	if (Char.sword >= sword_2_drawn) {
		// If char is holding sword, it makes redraw-area bigger.
		if (Char.direction >= dir_0_right) {
			if (++char_col_right > 9) char_col_right = 9;
			// char_col_right = MIN(char_col_right + 1, 9);
		} else {
			if (--char_col_left < 0) char_col_left = 0;
			// char_col_left = MAX(char_col_left - 1, 0);
		}
	}
	if (Char.charid == charid_0_kid) {
		x_top_row = MIN(char_top_row, prev_char_top_row);
		x_col_right = MAX(char_col_right, prev_char_col_right);
		x_col_left = MIN(char_col_left, prev_char_col_left);
	} else {
		x_top_row = char_top_row;
		x_col_right = char_col_right;
		x_col_left = char_col_left;
	}
	for (tile_row = x_top_row; tile_row <= char_bottom_row; ++tile_row) {
		for (tile_col = x_col_left; tile_col <= x_col_right; ++tile_col) {
			set_redraw_fore(get_tilepos(tile_col, tile_row), 1);
		}
	}
	if (Char.charid == charid_0_kid) {
		prev_char_top_row = char_top_row;
		prev_char_col_right = char_col_right;
		prev_char_col_left = char_col_left;
	}
}

// seg003:0645
void __pascal far redraw_at_char2() {
	short char_action;
	short char_frame;
	void __pascal (* redraw_func)(short, byte);
	char_action = Char.action;
	char_frame = Char.frame;
	redraw_func = &set_redraw2;
	// frames 78..80: grab
	if (char_frame < frame_78_jumphang || char_frame >= frame_80_jumphang) {
		// frames 135..149: climb up
		if (char_frame >= frame_137_climbing_3 && char_frame < frame_145_climbing_11) {
			redraw_func = &set_redraw_floor_overlay;
		} else {
			// frames 102..106: fall
			if (char_action != actions_2_hang_climb && char_action != actions_3_in_midair &&
			        char_action != actions_4_in_freefall && char_action != actions_6_hang_straight &&
			        (char_action != actions_5_bumped || char_frame < frame_102_start_fall_1 || char_frame > frame_106_fall)) {
				return;
			}
		}
	}
	for (tile_col = char_col_right; tile_col >= char_col_left; --tile_col) {
		if (char_action != 2) {
			redraw_func(get_tilepos(tile_col, char_bottom_row), 1);
		}
		if (char_top_row != char_bottom_row) {
			redraw_func(get_tilepos(tile_col, char_top_row), 1);
		}
	}
}

// seg003:0706
void __pascal far check_knock() {
	if (knock) {
		do_knock(Char.room, Char.curr_row - (knock>0));
		knock = 0;
	}
}

// seg003:0735
void __pascal far timers() {
	if (united_with_shadow > 0) {
		--united_with_shadow;
		if (united_with_shadow == 0) {
			--united_with_shadow;
		}
	}
	// custom logic - makes shadow blink on level 9
	if (kid_shadow_blink_timer > 0) {
		--kid_shadow_blink_timer;
	}
	if (guard_notice_timer > 0) {
		--guard_notice_timer;
		// custom logic - guard notice timer actually triggers the is_guard_notice event on levels except for the Jaffar's level
		if (guard_notice_timer == 0 && current_level < 14) {
			is_guard_notice = 1;
		}
	}
	if (resurrect_time > 0) {
		--resurrect_time;
	}
	// custom logic - level restart timer
	if (level_restart_timer > 0) {
		--level_restart_timer;
		if (level_restart_timer == 0) {
			if (current_level == 15) {
				start_level = -1;
				start_game();
				return;
			} else {
				is_restart_level = 1;
			}
		}
	}
	// custom logic blink timer
	if (blink_timer < 12) {
		++blink_timer;
		if (blink_timer == 12) {
			blink_timer = 0;
		}
	}
	if (is_feather_fall && !check_sound_playing()) {
#ifdef USE_REPLAY
		if (recording) special_move = MOVE_EFFECT_END;
		if (!replaying) // during replays, feather effect gets cancelled in do_replay_move()
#endif
			is_feather_fall = 0;
	}
	// Special event: mouse
	// custom logic - there is no delay for the mouse
	if (current_level == /*8*/ custom->mouse_level &&
			Kid.room == /*16*/ custom->mouse_room &&
			Kid.curr_row == 0 &&
			leveldoor_open == 2 &&
			get_tile(16, 3, 0) == tiles_4_gate &&
			curr_room_modif[curr_tilepos] >= 40) {
		++leveldoor_open;
		do_mouse();
	}
	// custom logic - mouse can be seen above above when kid has already joined shadow on level 9
	if (current_level == 12 &&
			initial_shadow_potion_counter == 0xFFFF &&
			Kid.room == 15 &&
			Kid.curr_row == 2 &&
			initial_mouse_timer >= 0 && 
			initial_mouse_timer < 2) {
		++initial_mouse_timer;
		if (initial_mouse_timer == 2) {
			initial_mouse_timer = -1;
			do_mouse_level12_1();
		}
	}
	// custom logic - mouse appears if kid has already joined shadow on level 9
	if (current_level == 12 &&
			initial_shadow_potion_counter == 0xFFFF &&
			Kid.room == 2 &&
			Kid.curr_row == 0 &&
			(Kid.curr_col == 8 || Kid.curr_col == 9) &&
			leveldoor_open < 72) {
		++leveldoor_open;
		if (leveldoor_open == 72) {
			play_sound(sound_41_end_level_music);
			do_mouse_level12_2();
		}
	}
	// custom logic - replace potion with an empty one
	if (potion_room > 0 && potion_pos >= 0 && potion_pos < 30) {
		if (++potion_counter > 20) {
			// calculating column and row from position
			int potion_col = -1;
			int potion_row = -1;
			int row_iterator;
			for (row_iterator = 2; row_iterator >= 0; row_iterator--) {
				if (potion_pos >= tbl_line[row_iterator]) {
					potion_row = row_iterator;
					potion_col = potion_pos - tbl_line[row_iterator];
					break;
				}
			}
			if (potion_col < 0 || potion_row < 0) {
				display_text_bottom("POTION POSITION ERROR");
				return;
			}
			get_tile(potion_room, potion_col, potion_row);
			curr_room_tiles[potion_pos] = tiles_10_potion;
			switch (potion_size) {
				case 1:
					curr_room_modif[potion_pos] = 0x40; // large empty potion
					break;
				case 2:
					curr_room_modif[potion_pos] = 0x30; // opener potion
					break;
				default:
					curr_room_modif[potion_pos] = 0x38; // small empty potion
					break;
			}
			redraw_height = 35;
			set_wipe(potion_pos, 1);
			set_redraw_full(potion_pos, 1);
			if (potion_size > 1) { // animate non-empty potions
				start_anim_potion(potion_room, potion_pos);
			}
			// special logic - visual effects after drinking a potion on level 6, room 24
			if (current_level == 6 && Char.room == 24) {
				if (potion_counter == 21) {
					// custom logic - level 6 has a different color palette
					load_lev6_palette();
					get_tile(Char.room, 0, 1);
					curr_room_tiles[curr_tilepos] = tiles_19_torch;
					redraw_height = 63;
					set_wipe(curr_tilepos, 1);
					set_redraw_full(curr_tilepos, 1);
					start_anim_torch(Char.room, curr_tilepos);
					is_blind_mode = 0;
					need_full_redraw = 1;
					flash_color = color_13_brightmagenta;
					flash_time = 20;
					stop_sounds();
					play_sound(sound_44_skel_alive);
				} else if (potion_counter >= 63) {
					display_text_bottom("WOAH!!! GROOVY!");
					text_time_remaining = 48;
					text_time_total = 48;
					potion_room = 0;
					potion_pos = -1;
					potion_counter = 0;
					potion_size = 0;
					is_show_time = 1;
				}
			} else {
				potion_room = 0;
				potion_pos = -1;
				potion_counter = 0;
				potion_size = 0;
			}
		}
	}
	// custom logic - remember whether the kid exited through the left door or not
	if (current_level == 2 && Kid.room == 23 && Kid.curr_row == 1) {
		level_2_right_door = Kid.curr_col >= 3 ? 1 : 0;
	}
	// custom logic - huge spilled poision potion kills kid without an antidote
	if (current_level == 5 &&
	        ((Kid.room == 15 && Kid.curr_col > 0 && Kid.x < 95) || Kid.room == 18) &&
	        Kid.curr_row == 1 &&
			Kid.alive < 0) {
		if (took_antidote == 0 || took_antidote == 1) {
			took_antidote = -1;
			display_text_bottom("FIND THE ANTIDOTE");
			text_time_total = 48;
			text_time_remaining = 48;
			flash_color = color_1_blue;
			flash_time = 5;
			loadkid();
			take_hp(100);
			stop_sounds();
			set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
			play_sound(sound_1_falling); // prince screaming
		} else if (took_antidote == 2) {
			++took_antidote;
			flash_color = color_13_brightmagenta;
			flash_time = 1;
		}
	}
	// custom logic - the kid must die to open level door
	if (current_level == 10 &&
			Kid.room == 8 &&
			Kid.curr_row == 2 &&
			(Kid.alive >= 0 || leveldoor_open >= 60) &&
			leveldoor_open <= 150) {
		++leveldoor_open;
		if (leveldoor_open == 60) {
			resurrect_time = 1;
			Kid.alive = -1;
			kid_shadow_blink_timer = 60;
			erase_bottom_text(1);
			stop_sounds();
			play_sound(sound_32_shadow_music);
		} else if (leveldoor_open == 150) {
			get_tile(8, 1, 0);
			trigger_button(0, tiles_15_opener, -1);
		}
	}
	if (current_level == 13) {
		// custom logic - standing near the large blue potion for too long kills kid
		if (Kid.room != 4 ||
		        hitp_curr == 0 ||
		        get_tile(4, 6, 0) != tiles_10_potion ||
		        curr_room_modif[get_tilepos(6, 0)] == 0x40) {
			kid_near_big_blue_potion = -1;
		} else if (kid_near_big_blue_potion >= 0 &&
		           kid_near_big_blue_potion < 301 &&
		           Kid.curr_col > 4 &&
		           Kid.curr_col < 8 &&
		           Kid.curr_row == 0) {
			if (++kid_near_big_blue_potion == 300) {
				// kill the kid
				kid_near_big_blue_potion = -1;
				flash_color = color_1_blue;
				flash_time = 2;
				loadkid();
				take_hp(100);
				stop_sounds();
				set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
				play_sound(sound_1_falling);
			}
		} else {
			kid_near_big_blue_potion = 0;
		}
		// custom logic - open left gate after the level door is opened on level 12B
		if (Kid.room == 1 && leveldoor_open == 1) {
			++leveldoor_open;
			// open the left gate
			get_tile(20, 0, 0);
			trigger_button(0, tiles_14_debris, -1);
		}
		// custom logic - kill kid after passing gate when the level door is closed on level 12B
		if (level_13_nowhere_to_run_timer >= 0 &&
		        Kid.room == 1 &&
		        ((Kid.curr_row == 0 && Kid.x >= 92) ||
		         Kid.alive >= 0)) {
			if (++level_13_nowhere_to_run_timer == 1 && Kid.alive < 0) {
				flash_color = color_1_blue;
				flash_time = 5;
				if (Kid.frame == frame_40_running_jump_1 && Kid.y < 55) {
					Kid.y = 60;
				} else if (Kid.frame == frame_41_running_jump_2 && Kid.y < 55) {
					Kid.y = 56;
				}
				loadkid();
				take_hp(100);
				stop_sounds();
				play_sound(sound_36_out_of_time);
			} else if (level_13_nowhere_to_run_timer > 36) {
				level_13_nowhere_to_run_timer = -1;
				Kid.charid = 5; // hides kid by setting his char id to a non-existing value
				int tile_col;
				if (get_tile(1, 4, 0) != tiles_21_skeleton) {
					curr_room_tiles[curr_tilepos] = tiles_21_skeleton;
					redraw_height = 35;
					set_wipe(curr_tilepos, 1);
					set_redraw_full(curr_tilepos, 1);
					set_wipe(curr_tilepos + 1, 1);
					set_redraw_full(curr_tilepos + 1, 1);
					if (Kid.curr_row == 0) {
						for (tile_col = 1; tile_col <= 5; tile_col++) {
							get_tile(1, tile_col, 1);
							curr_room_tiles[curr_tilepos] = tiles_10_potion;
							redraw_height = 63;
							set_wipe(curr_tilepos, 1);
							set_redraw_full(curr_tilepos, 1);
							set_redraw_anim_curr();
							set_wipe(curr_tilepos + 1, 1);
							set_redraw_full(curr_tilepos + 1, 1);
							start_anim_potion(1, curr_tilepos);
						}
					}
				}
			}
		}
	}
	// custom logic - on level 13 open left most gate after Jaffar dies
	if (current_level == 14 && Kid.room == 1 && leveldoor_open == 2) {
		have_sword = false;
		++leveldoor_open;
		get_tile(8, 1, 0);
		trigger_button(0, 0, -1);
	}
}

// seg003:0798
void __pascal far check_mirror() {
	word clip_top;
	if (jumped_through_mirror == -1) {
		jump_through_mirror();
	} else {
		if (get_tile_at_char() == tiles_13_mirror) {
			loadkid();
			load_frame();
			check_mirror_image();
			if (distance_mirror >= 0 && custom->show_mirror_image) {
				load_frame_to_obj();
				reset_obj_clip();
				clip_top = y_clip[Char.curr_row + 1];
				if (clip_top < obj_y) {
					obj_clip_top = clip_top;
					obj_clip_left = (Char.curr_col << 5) + 9;
					add_objtable(4); // mirror image
				}
			}
		}
	}
}

// seg003:080A
void __pascal far jump_through_mirror() {
	// custom logic - on level 6, room 1 change tiles after the jump so kid would fall down
	if (current_level == 6) {
		int tile_row, tile_col;
		for (tile_row = 2; tile_row >= 1; tile_row--)
			for (tile_col = 2; tile_col <= 6; tile_col++) {
				get_tile(1, tile_col, tile_row);
				if (tile_row == 2 && tile_col < 5) {
					curr_room_tiles[curr_tilepos] = tiles_0_empty;
					curr_room_modif[curr_tilepos] = 1;
				} else if (tile_row == 1 && tile_col == 5) {
					curr_room_tiles[curr_tilepos] = tiles_3_pillar;
				}
				redraw_height = 63;
				set_wipe(curr_tilepos, 1);
				set_redraw_full(curr_tilepos, 1);
			}
		get_tile(1, 6, 1);
		curr_room_tiles[curr_tilepos] = tiles_15_opener;
		curr_room_modif[curr_tilepos] = 23;
		get_tile(1, 0, 1);
		trigger_button(0, tiles_15_opener, -1);
	}
	loadkid();
	load_frame();
	check_mirror_image();
	jumped_through_mirror = 0;
	Char.charid = charid_1_shadow;
	set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
	play_sound(sound_45_jump_through_mirror); // jump through mirror
	saveshad();
	draw_rect(&rect_bottom_text, 0);
	if (current_level == 6) {
		guardhp_max = guardhp_curr = 1;
		draw_guard_hp(guardhp_curr, guardhp_max);
		clear_kid_hp(hitp_max);
		draw_kid_hp(MIN(hitp_curr, hitp_max - 1), hitp_max - 1);
	}
	// custom logic - flash red when jumping through mirror
	flash_color = 12; // red
	flash_time = 10;
}

// seg003:085B
void __pascal far check_mirror_image() {
	short distance;
	short xpos;
	xpos = x_bump[Char.curr_col + 5] + 10;
	distance = distance_to_edge_weight();
	if (Char.direction >= dir_0_right) {
		distance = (~distance) + 14;
	}
	distance_mirror = distance - 2;
	Char.x = (xpos << 1) - Char.x;
	Char.direction = ~Char.direction;
}

// seg003:08AA
void __pascal far bump_into_opponent() {
	// This is called from play_kid_frame, so char=Kid, Opp=Guard
	short distance;
	if (can_guard_see_kid >= 2 &&
	        Char.sword == sword_0_sheathed && // Kid must not be in fighting pose
	        Opp.sword != sword_0_sheathed && // but Guard must
	        Opp.action < 2 &&
	        Char.direction != Opp.direction // must be facing toward each other
	   ) {
		distance = char_opp_dist();
		if (ABS(distance) <= 15) {

#ifdef FIX_PAINLESS_FALL_ON_GUARD
			if (fixes->fix_painless_fall_on_guard) {
				if (Char.fall_y >= 33) return; // don't bump; dead
				else if (Char.fall_y >= 22) { // medium land
					take_hp(1);
					set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
					play_sound(sound_16_medium_land);
				}
			}
#endif

			Char.y = y_land[Char.curr_row + 1];
			Char.fall_y = 0;
			seqtbl_offset_char(seq_47_bump); // bump into opponent
			play_seq();
		}
	}
}

// seg003:0913
void __pascal far pos_guards() {
	// Custom logic - add a guard to level 15
	if (current_level == 15) {
		level.guards_tile[0] = 7;
		level.guards_dir[0] = dir_FF_left;
		level.guards_color[0] = 2;
		level.guards_skill[0] = 1;
	}

	short guard_tile;
	short room1;
	for (room1 = 0; room1 < 24; ++room1) {
		guard_tile = level.guards_tile[room1];
		if (guard_tile < 30) {
			level.guards_x[room1] = x_bump[guard_tile % 10 + 5] + 14;
			level.guards_seq_hi[room1] = 0;
		}
	}
}

// seg003:0959
void __pascal far check_can_guard_see_kid() {
	/*
	Possible results in can_guard_see_kid:
	0: Guard can't see Kid
	1: Guard can see Kid, but won't come
	2: Guard can see Kid, and will come
	*/
	short kid_frame;
	short left_pos;
	short temp;
	short right_pos;
	kid_frame = Kid.frame;
	if (Guard.charid == charid_24_mouse) {
		return;
	}
	if ((Guard.charid != charid_1_shadow || current_level == 12) &&
	        // frames 217..228: going up on stairs
	        kid_frame != 0 && (kid_frame < frame_219_exit_stairs_3 || kid_frame >= 229) &&
	        Guard.direction != dir_56_none && Kid.alive < 0 && Guard.alive < 0 && Kid.room == Guard.room && Kid.curr_row == Guard.curr_row
	   ) {
		can_guard_see_kid = 2;
		left_pos = x_bump[Kid.curr_col + 5] + 7;
		right_pos = x_bump[Guard.curr_col + 5] + 7;
		if (left_pos > right_pos) {
			temp = left_pos;
			left_pos = right_pos;
			right_pos = temp;
		}
		// A chomper is on the left side of a tile, so it doesn't count.
		if (get_tile_at_kid(left_pos) == tiles_18_chomper) {
			left_pos += 14;
		}
		// A gate is on the right side of a tile, so it doesn't count.
		if (get_tile_at_kid(right_pos) == tiles_4_gate) {
			right_pos -= 14;
		}
		// A guard standing on a door top (with floor) should not become inactive.
		if (get_tile_at_kid(right_pos) == tiles_7_doortop_with_floor || get_tile_at_kid(right_pos) == tiles_12_doortop) {
			right_pos -= 14;
		}
		if (right_pos >= left_pos) {
			while (left_pos <= right_pos) {
				// Can't see through these tiles.
				if (get_tile_at_kid(left_pos) == tiles_20_wall ||
				        curr_tile2 == tiles_7_doortop_with_floor ||
				        curr_tile2 == tiles_12_doortop
				   ) {
					can_guard_see_kid = 0;
					return;
				}
				// Can see through these, but won't go through them.
				if (curr_tile2 == tiles_11_loose ||
				        curr_tile2 == tiles_18_chomper ||
				        (curr_tile2 == tiles_4_gate && curr_room_modif[curr_tilepos] < 112) ||
				        !tile_is_floor(curr_tile2)
				   ) {
					can_guard_see_kid = 1;
				}
				left_pos += 14;
			}
		}
	} else {
		can_guard_see_kid = 0;
	}
}

// seg003:0A99
byte __pascal far get_tile_at_kid(int xpos) {
	return get_tile(Kid.room, get_tile_div_mod_m7(xpos), Kid.curr_row);
}

// seg003:0ABA
void __pascal far do_mouse() {
	loadkid();
	Char.charid = /*charid_24_mouse*/ custom->mouse_object;
	Char.x = /*200*/ custom->mouse_start_x;
	Char.curr_row = 0;
	Char.y = y_land[/*Char.curr_row*/ 0 + 1];
	Char.alive = -1;
	Char.direction = dir_FF_left;
	Char.fall_x = 0;
	Char.fall_y = 0;
	guardhp_curr = 1;
	seqtbl_offset_char(seq_105_mouse_forward); // mouse forward
	play_seq();
	saveshad();
}

// custom logic - mouse is first visible above on level 12 if kid joined shadow on level 9
void __pascal far do_mouse_level12_1() {
	loadkid();
	Char.charid = /*charid_24_mouse*/ custom->mouse_object;
	Char.x = 70;
	Char.curr_row = 0;
	Char.y = y_land[Char.curr_row + 1];
	Char.alive = -1;
	Char.direction = dir_FF_left;
	Char.fall_x = 0;
	Char.fall_y = 0;
	guardhp_curr = 1;
	seqtbl_offset_char(seq_101_mouse_stands_up);
	play_seq();
	saveshad();
}

// custom logic - mouse appears on level 12 if kid joined shadow on level 9
void __pascal far do_mouse_level12_2() {
	loadkid();
	Char.charid = /*charid_24_mouse*/ custom->mouse_object;
	Char.x = /*200*/ custom->mouse_start_x;
	Char.curr_row = 0;
	Char.y = y_land[Char.curr_row + 1];
	Char.alive = -1;
	Char.direction = dir_FF_left;
	Char.fall_x = 0;
	Char.fall_y = 0;
	guardhp_curr = 1;
	seqtbl_offset_char(seq_105_mouse_forward); // mouse forward
	play_seq();
	saveshad();
}

// seg003:0AFC
int __pascal far flash_if_hurt() {
	if (flash_time != 0) {
		do_flash(flash_color);
		return 1;
	} else if (hitp_delta < 0) {
		if (is_joyst_mode && enable_controller_rumble) {
			if (sdl_haptic != NULL) {
				SDL_HapticRumblePlay(sdl_haptic, 1.0, 100); // rumble at full strength for 100 milliseconds
#if SDL_VERSION_ATLEAST(2,0,9)
			} else if (sdl_controller_ != NULL) {
				SDL_GameControllerRumble(sdl_controller_, 0xFFFF, 0xFFFF, 100);
			} else {
				SDL_JoystickRumble(sdl_joystick_, 0xFFFF, 0xFFFF, 100);
#endif
			}
		}
		do_flash(color_12_brightred); // red
		return 1;
	}
	return 0; // not flashed
}

// seg003:0B1A
void __pascal far remove_flash_if_hurt() {
	if (flash_time != 0) {
		--flash_time;
	} else {
		if (hitp_delta >= 0) return;
	}
	remove_flash();
}
